{-$Define AskPowerChange}
{$IFDEF VirtualPascal}
  {&Use32+}
  {&Delphi-}
  {&AlignCode+}
  {&AlignData+}
  {&AlignRec-}
  {&Optimize+}
  {$M 40000}
{$ELSE}
  {$M 30000,0,0}
{$ENDIF}

Program PCI;

{$G+}
{$R+}
{$S-}
{$I+}
{$N-}
{$E-}


uses
  {$IfDef VirtualPascal}
  VpUtils,
  {$Else}
  {newdelay, - we do not Delay}
  {$EndIf}
  WinDos,Dos,
  Crt,
  Strings,
  pci_hw,
  redircon;


{$I classes.pas}


{
  This code is Written by Craig Hart in 1996-2001. It is released as freeware;
  please use and modify at will. No gurarantees are made or implied.


  Please read the accompaning documentation PCI.DOC for all the info
  relating to this program!
}



const
  revision      ='0.48vkE';
  FModeReadOnly = $40; (* open_share_DenyNone+open_access_ReadOnly *)

var
  wrlncount,
  deviceid,
  func,
  nn,
  pp,
  lb,
  bus,
  sum,
  disp,
  cap_ptr       : byte;

  found,
  businfo,
  tableok,
  dorouting,
  dopcirouting,
  userev,
  summary,
  bogusid,
  genssid,
  dumpregs,
  first,
  installermode,
  showhelp,
  vga50         : boolean;


  irqmap        : array[byte] of byte;


  romsize,
  org_rom_addr  : longint;



  addr          : longint;

  i,
  j,
  l,
  v             : word;

  f             : text;

  revchk,
  oemidnum,
  oemidstr,
  cmdstr,
  vstr,
  cmpstr        : string;

  pcidevs_txt   : string;
  pcidevs_path  : string;

  infotbl_cache : array[0..$ff] of byte;
  infotbl_read  : array[0..$ff] of boolean;
  infotbl_deviceid,
  infotbl_func,
  infotbl_bus   : word;

  bustype       : array[0..255] of (bustype_PCI_known,bustype_PCI,bustype_AGP,bustype_CardBus,bustype_None);


(* Read-Cache functions for PCI register space
   replaces old array infotbl - no need to read ahead and hang.. *)

procedure reset_infotbl_cache;
  begin
    FillChar(infotbl_read,SizeOf(infotbl_read),false);
  end;

procedure set_lookup_device(const deviceid,func,bus:word);
  begin
    infotbl_deviceid:=deviceid;
    infotbl_func:=func;
    infotbl_bus:=bus;
    reset_infotbl_cache;
  end;

function infotbl(const i:word):byte;
  begin
    if not infotbl_read[i] then
      begin
        infotbl_cache[i]:=lookup(infotbl_deviceid,infotbl_func,infotbl_bus,i);
        infotbl_read[i]:=true;
      end;
    infotbl:=infotbl_cache[i];
  end;

function infotbl_W(const i:word):Word;
  begin
    infotbl_W:=Word(infotbl(i+1)) shl 8
              +     infotbl(i  )       ;
  end;

function infotbl_L(const i:word):longint;
  begin
    infotbl_L:=Longint(infotbl_W(i+2)) shl 16
              +        infotbl_W(i  )        ;
  end;


function cvt4(const b:byte) : char;
  begin
    if b>9 then cvt4:=Chr(b+Ord('A')-10)
    else cvt4:=Chr(b+Ord('0'));
  end;

function wrhexb(const byt:byte): string;
  begin
    wrhexb:=cvt4(byt and $0f);
  end;

function wrhex(const byt:byte) : string;
  begin
    wrhex[0]:=#2;
    wrhex[1]:=cvt4((byt shr 4) and $0f);
    wrhex[2]:=cvt4( byt        and $0f);
  end;

function wrhexw(const wor:word): string;
  begin
    wrhexw[0]:=#4;
    wrhexw[1]:=cvt4((wor shr 12) and $0f);
    wrhexw[2]:=cvt4((wor shr  8) and $0f);
    wrhexw[3]:=cvt4((wor shr  4) and $0f);
    wrhexw[4]:=cvt4( wor         and $0f);
  end;

function wrhexl(const lon:longint): string;
  begin
    wrhexl:=wrhexw(lon shr 16)
           +wrhexw(lon and $ffff);
  end;


(* Make the PCI configuration status register printout pretty *)
(* Input = the string to be output *)

procedure printstatus (const s : string);
  begin
    if not first then
      if (Length(s)+WhereX)>78 then
        begin
          WriteLn(',');
          Write('   ');
        end
      else
        Write(', ');
    Write(s);
    first:=false;
  end;

procedure WriteColor(const s: string; a: byte);
  var
    oa: Byte;
  begin
    oa := TextAttr;
    TextAttr := (TextAttr and (not $0f)) or a;
    Write(s);
    TextAttr := oa;
  end;

procedure WritePCColor(const s: pChar; a: byte);
  var
    oa: Byte;
  begin
    oa := TextAttr;
    TextAttr := (TextAttr and (not $0f)) or a;
    Write(s);
    TextAttr := oa;
  end;

procedure WriteLnColor(const s: string; a: byte);
  begin
    WriteColor(s,a);
    WriteLn;
  end;

{$IfDef VirtualPascal}
var
  pcidevs_txt_buffer    :PChar;
  pcidevs_txt_end       :PChar;
  pcidevs_txt_position  :PChar;
{$EndIf}

procedure Assign2(var f:text;const filename:string);
  {$IfDef VirtualPascal}
  var
    f2                  :file;
    f2s                 :longint;

  begin
    pcidevs_txt_buffer  :=nil;
    pcidevs_txt_end     :=nil;
    pcidevs_txt_position:=nil;

    FileMode:=FModeReadOnly;
    Assign(f2,filename);
    {$I-}
    Reset(f2,1);
    {$I+}
    if InOutRes<>0 then Exit;

    f2s:=FileSize(f2);
    GetMem(pcidevs_txt_buffer,f2s+3);
    pcidevs_txt_buffer[f2s  ]:=#$0d;
    pcidevs_txt_buffer[f2s+1]:=#$0a;
    pcidevs_txt_buffer[f2s+2]:=#$00;
    BlockRead(f2,pcidevs_txt_buffer[0],f2s);
    Close(f2);

    if not installermode then
      begin
        (* '; This is version 474 of this file; 04-06-2004 (D-M-Y).' *)
        pcidevs_txt_position:=StrPos(pcidevs_txt_buffer,#$0a'; This is version ');
        if Assigned(pcidevs_txt_position) then
          begin
            Inc(pcidevs_txt_position,11-8);
            Write('Loaded Vendor Data : "');
            while pcidevs_txt_position[0]>=' ' do
              begin
                Write(pcidevs_txt_position[0]);
                Inc(pcidevs_txt_position);
              end;
            WriteLn('"');
          end
        else
          WriteLnColor('Note: Please download a complete Vendor Data file (pcidevs.txt)!',LightRed);
      end;

    pcidevs_txt_end     :=@pcidevs_txt_buffer[f2s];
    pcidevs_txt_position:=pcidevs_txt_buffer;

  end;
  {$Else}
  begin
    FileMode:=FModeReadOnly;
    Assign(f,filename);
  end;
  {$EndIf}

{$IfNDef VirtualPascal}
var
  pcidevs_txt_read_buffer:array[0..32*1024-1] of byte;
{$EndIf}

procedure Reset2(var f:text);
  {$IfDef VirtualPascal}
  begin
    pcidevs_txt_position:=pcidevs_txt_buffer;
  end;
  {$Else}
  begin
    Reset(f);
    SetTextBuf(f,pcidevs_txt_read_buffer,SizeOf(pcidevs_txt_read_buffer));
  end;
  {$EndIf}


procedure ReadLn2(var f:text;var zk:string);
  {$IfDef VirtualPascal}
  var
    linestart:PChar;
    linelength:word;
  begin

    linestart:=pcidevs_txt_position;
    linelength:=0;
    while pcidevs_txt_position<>pcidevs_txt_end do
      case pcidevs_txt_position[0] of
        ^Z:
          Break;
        #13:
          Inc(pcidevs_txt_position);

        #10:
          begin
            Inc(pcidevs_txt_position);
            Break;
          end;
      else
        Inc(pcidevs_txt_position);
        Inc(linelength);
      end;

    if linelength=0 then
      zk:=''
    else
      begin
        SetLength(zk,linelength);
        Move(linestart^,zk[1],linelength);
      end;

  end;
  {$Else}
  begin
    ReadLn(f,zk);
  end;
  {$EndIf}


function Eof2(var f:text):boolean;
  {$IfDef VirtualPascal}
  begin
    Eof2:=(pcidevs_txt_position=pcidevs_txt_end) or (pcidevs_txt_position^=^Z);
  end;
  {$Else}
  begin
    Eof2:=Eof(f);
  end;
  {$EndIf}


procedure Close2(var f:text);
  {$IfDef VirtualPascal}
  begin
  end;
  {$Else}
  begin
    Close(f);
  end;
  {$EndIf}


procedure listmap(const va:word;const dispst:string);
  var
    comma  : byte;
    failed : boolean;
    l,
    j      : word;

  begin
    failed:=true;
    Write(dispst);
    comma:=0;
    for l:=0 to 15 do
      Inc(comma,(va shr l) and 1);

    for l:=0 to 15 do
      if Odd(va shr l) then
        begin
          Write(l);
          if comma>1 then Write(',')
          else Write(' ');
          Dec(comma);
          failed:=false;
        end;

    if failed then
      WriteLn('None')
    else
      WriteLn;
  end;

procedure listmap2(const va_is,va_should:word;const dispst:string);
  var
    comma  : byte;
    failed : boolean;
    l,
    j      : word;

  begin
    failed:=true;
    Write(dispst);
    comma:=0;
    for l:=0 to 15 do
      Inc(comma,(va_is shr l) and 1);

    for l:=0 to 15 do
      if Odd(va_is shr l) then
        begin
          if Odd(va_should shr l) then
            TextColor(LightGray)
          else
            TextColor(LightRed);
          Write(l);
          TextColor(LightGray);
          if comma>1 then Write(',')
          else Write(' ');
          Dec(comma);
          failed:=false;
        end;

    if failed then
      WriteLn('None')
    else
      WriteLn;
  end;


procedure lookupven(const silent:boolean);
  begin
    Reset2(f);
    failed:=true;
    repeat
      ReadLn2(f,vstr);
      if (vstr[1]='V') and (StrLComp(@vstr[3],@cmpstr[1],4)=0) then
        begin
          if not silent then WriteColor(Copy(vstr,8,Length(vstr)),Yellow);
          failed:=false;
        end;
    until Eof2(f) or not failed;

    if failed then
      if not silent then WriteColor('Unknown',LightRed);
  end;

procedure lookupdev(const LengthLimit:word);
  begin
    failed:=true;

    while failed and (not Eof2(f)) do
      begin

        ReadLn2(f,vstr);
        if (vstr[1]='D') and (StrLComp(@vstr[3],@cmpstr[1],4)=0) then
          begin
            if not Eof2(f) then ReadLn2(f,revchk);
            if revchk[1]='R' then
            begin
              repeat
                if wrhex(infotbl(8))=Copy(revchk,3,2) then vstr:='xxxxxxx'+Copy(revchk,6,Length(revchk));
                if not Eof2(f) then ReadLn2(f,revchk);
              until revchk[1]<>'R';
            end;
            Delete(vstr,1,1+1+4+1);
            if Length(vstr)>LengthLimit then
              begin
                SetLength(vstr,LengthLimit-2);
                vstr:=vstr+'..';
              end;
            WriteColor(vstr,Yellow);
            failed:=false;
          end;

      if vstr[1]='V' then Break;
    end;

    if failed then
      WriteColor('Unknown',LightRed);

  end;


procedure showinstallerinfo;
  begin
    Write('V:',wrhexw(infotbl_W(0)),' ');

    Write('D:',wrhexw(infotbl_W(2)),' ');

    Write('S:');
    case infotbl($e) and $7f of
      0:
        begin
          Write(wrhexw(infotbl_W($2e)));
          Write(wrhexw(infotbl_W($2c)),' ');
        end;
      2:
        begin
          Write(wrhexw(infotbl_W($42)));
          Write(wrhexw(infotbl_W($40)),' ');
        end;

    else
      Write('00000000 ');
    end;

    Write('B:',bus,' ');

    Write('E:');
    if deviceid<10 then Write('0'); (* 00..31 *)
    Write(deviceid,' ');

    Write('F:',func,' ');

    Write('I:'); (* Interrupt line *)
    if infotbl($3c) in [1..254] then
      begin
        if infotbl($3c)<10 then Write('0');
        Write(infotbl($3c),' ');
      end
    else
      Write('00 ');

    Write('N:'); (* Interrupt pin *)
    {if infotbl($3c) in [1..254] then
      begin
        if infotbl($3d)=0 then Write('- ')
        else Write(Chr(infotbl($3d)+Ord('@')),' ');
      end
    else Write('- ');}
    if infotbl($3d) in [1..4] then Write(Chr(infotbl($3d)+Ord('@')),' ')
    else                           Write('- ');

    Write('C:',wrhex(infotbl($b)),' ');

    Write('U:',wrhex(infotbl($a)),' ');

    Write('P:',wrhex(infotbl($9)),' ');

    write('R:',wrhex(infotbl($8)),' ');

    WriteLn;
  end;

const
  verify_routing_table_failed : boolean = false;

function verify_routing_table(const table:TIRQ_routing_table_entry_Array;slots:integer):boolean;
  var
    link_data           :array[byte] of word;
    link_data_initialized:array[byte] of integer;
    icbm1,icbm2         :word;
    i,j                 :integer;
  begin
    (* previous call displayed errors? *)
    if verify_routing_table_failed then
      begin
        verify_routing_table:=false;
        Exit;
      end;

    for i:=0 to slots-1-1 do
      with table[i] do
        if (PCI_bus_number<>0) or (PCI_device_number<>0) or (device_slot_number<>0) then
          for j:=i+1 to slots-1 do
            if   (table[j].PCI_bus_number         =PCI_bus_number         )
            and  (table[j].PCI_device_number shr 3=PCI_device_number shr 3) then
              begin
                verify_routing_table_failed:=true;
                TextColor(LightRed);
                WriteLn('  Error: table entries ',i,' and ',j,' refer to the same Bus=',
                  PCI_bus_number,'/Device=',PCI_device_number shr 3,'!');
                TextColor(LightGray);
              end;

    FillChar(link_data_initialized,SizeOf(link_data_initialized),$ff);
    for i:=0 to slots-1 do
      with table[i] do
        for j:=Low(INTABCD) to High(INTABCD) do
          with INTABCD[j] do
            begin
              if link_data_initialized[link_value]=-1 then
                begin
                  link_data[link_value]:=IRQ_connectivity_bit_map;
                  link_data_initialized[link_value]:=i*4+j;
                end
              else
              if link_value<>0 then
                begin
                  if link_data[link_value]<>IRQ_connectivity_bit_map then
                    begin
                      verify_routing_table_failed:=true;
                      TextColor(LightRed);
                      WriteLn('  Error: connectivity mismatches for link value ',wrhex(link_value),':');
                      TextColor(LightGray);

                      icbm1:=IRQ_connectivity_bit_map;

                      with table[link_data_initialized[link_value] div 4] do
                      with INTABCD[link_data_initialized[link_value] mod 4] do
                      begin
                      Write  ('    Bus=',PCI_bus_number,' Device=',PCI_device_number shr 3,' Pin=#',Chr(Ord('@')+j+1),' Map=');
                      icbm2:=IRQ_connectivity_bit_map;
                      TextColor(LightRed);
                      (*listmap(IRQ_connectivity_bit_map,'');*)
                      listmap2(icbm2,icbm1,'');
                      TextColor(LightGray);
                      end;
                      Write  ('    Bus=',PCI_bus_number,' Device=',PCI_device_number shr 3,' Pin=#',Chr(Ord('@')+j+1),' Map=');
                      TextColor(LightRed);
                      (*listmap(IRQ_connectivity_bit_map,'');*)
                      listmap2(icbm1,icbm2,'');
                      TextColor(LightGray);
                    end;
                end;
            end;
     verify_routing_table:=not verify_routing_table_failed;
  end;


procedure showroutinginfo;
  begin
    (*WriteLn('ROM PCI IRQ routing table Windows 9x Compatibility Tests....');*)
    WriteLn('ROM PCI IRQ routing table Tests....');

{ Find table }

    i:=0;
    failed:=true;
    repeat
      if MemL_F000(i)=$52495024 (* $PIR *) then failed:=false;
      if failed then Inc(i,$10);
    until (i>$ffef) or (not failed);

{ check table }

    if failed then
      begin
        WriteLnColor('No ROM PCI IRQ routing table found!!!',LightRed);
        Exit;
      end;

    tableok:=true;


    WriteLn(' ROM IRQ routing table found at F000h:',wrhexw(i),'h');
    Write(' Table Version ',Mem_F000(i+5),'.',Mem_F000(i+4));
    if (Mem_F000(i+5)=1) and (Mem_F000(i+4)=0) then
      WriteLn(' - OK')
    else
      begin
        WriteLnColor('Invalid Version!',LightRed);
        tableok:=false;
      end;

    Write(' Table size ',MemW_F000(i+6),' bytes - ');
    if (MemW_F000(i+6)<33) or (MemW_F000(i+6) mod 16<>0) then
      begin
        WriteLnColor('Invalid Size!',LightRed);
        tableok:=false;
      end
    else
      WriteLn('OK');


    if tableok then
      begin
        Write(' Table Checksum ',wrhex(Mem_F000(i+31)),'h - ');
        {$R-}  {Range checking off as sum is DELIBERATELY meant to overfow }
        sum:=0;
        for l:=0 to MemW_F000(i+6)-1 do
          Inc(sum,Mem_F000(i+l));
        {$R+}
        if sum=0 then
          WriteLn('OK')
        else
          begin
            WriteLnColor('Failed!',LightRed);
            tableok:=false;
          end;
      end;

    listmap(MemW_F000(i+10),' IRQ''s dedicated to PCI : ');

    if tableok then
      tableok:=verify_routing_table(TIRQ_routing_table_entry_Array(Ptr_F000(i+$20)^),(MemW_F000(i+6)-$20) div $10);

    if tableok then
      WriteLnColor(' The ROM PCI IRQ routing table appears to be OK.',LightGreen)
    else
      WriteLnColor(' The ROM PCI IRQ routing table appears to be faulty!!',LightRed);

  end;


procedure dohexdump;
  begin
    WriteLn('Hex-Dump of IRQ Routing table : ');
    WriteLn;

    Write('  0000  ');

    for j:=0 to 15 do
      if j<6 then
        Write(wrhex(irqbuff[j]),' ')
      else
        Write('   ');

    Write('   ');

    for j:=0 to 6-1 do
      if irqbuff[j]<32 then
        Write('.')
      else
        Write(Chr(irqbuff[j]));

    WriteLn;


    for i:=Low(irqbuffR.IRQ_routing_table_entry_Array) to High(irqbuffR.IRQ_routing_table_entry_Array) do
      with irqbuffR.IRQ_routing_table_entry_Array[i] do
        begin
          Write('  ',wrhexw(i*16+6),'  ');

          for j:=0 to 15 do
            Write(wrhex(binary[j]),' ');

          Write('   ');

          for j:=0 to 15 do
            if binary[j]<32 then
              Write('.')
            else
              Write(Chr(binary[j]));

          WriteLn;
        end;

    WriteLn;
    WriteLn;
  end;


procedure docapdecode;
  const
    YesNo       :array[boolean] of string[3]=('No','Yes');
    Aux_Current :array[0..7] of word=(0,55,100,160,220,270,320,375);

  var
    xx          :byte;
    reply       :string;

  begin
    writeln(' New Capabilities List Present:');

    case infotbl($e) and $7f of
      0:cap_ptr:=infotbl($34);
      1:cap_ptr:=infotbl($34);
      2:cap_ptr:=infotbl($14);
    else
        cap_ptr:=0;
    end;

    if cap_ptr=0 then
      begin
        WriteLn('  No ''New Capabilities'' Are Currently Enabled');
        Exit;
      end;


    repeat

      if cap_ptr<$40 then
        begin
          WriteColor('Cap_Ptr is invalid!', LightRed);
          Break;
        end;

      case infotbl(cap_ptr) of

        01 :
          begin
            Write('   Power Management Capability'); (* "PMC", pm1_1.pdf page 25 *)

            (* registers:
               00 (B) ID(1)
               01 (B) next pointer
               02 (W) power management capabilities
               04 (W) power management control/status
               06 (B) bridge support extensions
               07 (B) data                              *)

            case (infotbl(cap_ptr+2) and 7) of
              0:Write(', Version before 1.0?');
              1:Write(', Version 1.0');
              2:Write(', Version 1.1');
            else
               Write(', unknown Version (later than 1.1)');
            end;
            WriteLn;
            if Odd(infotbl_w(cap_ptr+2) shr 10) then WriteLn('     Supports power state D2');
            if Odd(infotbl_w(cap_ptr+2) shr  9) then WriteLn('     Supports power state D1');
            if (infotbl_w(cap_ptr+2) shr 6) and 7<>0 then
              WriteLn('     current requirement for 3.3Vaux: ',Aux_Current[(infotbl_w(cap_ptr+2) shr 6) and 7],' mA');

            Write('     Current Power State : D');
            case infotbl(cap_ptr+4) and 3 of
              0 : WriteLn('0 (Device operational, no power saving)');
              1 : WriteLn('1 (Device idle, minimum power saving)');
              2 : WriteLn('2 (Device CLK stopped, medium power saving)');
              3 : WriteLn('3 (Device off, no power to device)');
            end;

            {$IfDef AskPowerChange}
            if not ioredirected then
              begin
                Write('New value (0=on 1=idle 2=medium 3=off) ');
                xx:=infotbl(cap_ptr+4) and (not 3);
                ReadLn(reply);
                if reply='0' then
                  write_byte(infotbl_deviceid,infotbl_func,infotbl_bus,cap_ptr+4,xx or 0)
                else
                if reply='1' then
                  write_byte(infotbl_deviceid,infotbl_func,infotbl_bus,cap_ptr+4,xx or 1)
                else
                if reply='2' then
                  write_byte(infotbl_deviceid,infotbl_func,infotbl_bus,cap_ptr+4,xx or 2)
                else
                if reply='3' then
                  write_byte(infotbl_deviceid,infotbl_func,infotbl_bus,cap_ptr+4,xx or 3);
                reset_infotbl_cache;
                Write(' New Current Power State : D');
                case infotbl(cap_ptr+4) and 3 of
                  0 : WriteLn('0 (Device operational, no power saving)');
                  1 : WriteLn('1 (Device idle, minimum power saving)');
                  2 : WriteLn('2 (Device CLK stopped, medium power saving)');
                  3 : WriteLn('3 (Device off, no power to device)');
                end;
              end;
            {$EndIf AskPowerChange}

            (* D0/1/2/3-state power consumed in watts *)
            for i:=0 to 8 do
              begin
                (* bit 15(7): PME status, write 0 has no effect
                   bit 14..13(6..5) data scale, read only
                   bit 12..09(4..1) data select
                   bit 8(0) PME enable (R/W) *)
                addr:=(infotbl(cap_ptr+4+1) and $01)
                     +i shl 1;
                write_byte(deviceid,func,bus,cap_ptr+4+1,addr);
                reset_infotbl_cache;
                case (infotbl_W(cap_ptr+4) shr 13) and 3 of
                  0:(*Write('???  ');*)
                    Continue; (* Data/_Select/_Scale not implemented *)
                  1:Write('     ',infotbl(cap_ptr+7)*100:5);
                  2:Write('     ',infotbl(cap_ptr+7)* 10:5);
                  3:Write('     ',infotbl(cap_ptr+7)*  1:5);
                end;
                Write(' mW ');
                case i of
                  0,1,2,3:Write('D',i  ,' power consume');
                  4,5,6,7:Write('D',i-4,' power dissipate');
                  8:      Write('common logic power consume');
                end;
                WriteLn;
              end;

          end;

        02 :
          begin
            Write('   AGP Capability, Version ');
            Write(infotbl(cap_ptr+2) shr 4,'.',infotbl(cap_ptr+2) and $0f,' ');

            if (infotbl(cap_ptr+2) shr 4)=1 then WriteLn('(AGP 1x and/or 2x support)');
            if (infotbl(cap_ptr+2) shr 4)=2 then WriteLn('(AGP 4x and below support)');
            if (infotbl(cap_ptr+2) shr 4)=3 then
              begin
                write('(AGP 8x and below');
                case infotbl(cap_ptr+2) and $0f of
                  0..4:Write(', core register');
                  5..9:Write(', appendix register');
                end;
              WriteLn(' support)');
            end;


            { Status register }

            Write('     AGP Speed(s) Supported : ');

            if infotbl(cap_ptr+4) and 8=8 then
              begin
                if infotbl(cap_ptr+4) and 1=1 then Write('4x ');
                if infotbl(cap_ptr+4) and 2=2 then Write('8x ');
                if infotbl(cap_ptr+4) and 4=4 then Write('Unknown Speed Reported! ');
              end
            else
              begin
                if infotbl(cap_ptr+4) and 1=1 then Write('1x ');
                if infotbl(cap_ptr+4) and 2=2 then Write('2x ');
                if infotbl(cap_ptr+4) and 4=4 then Write('4x ');
                if infotbl(cap_ptr+4) and 7=0 then
                  begin
                    WriteColor('None!!',LightRed);
                    WriteColor(' (Assume Only 1x Support)',LightCyan);
                  end;
              end;

            WriteLn;

            WriteLn('     FW Transfers Supported : ',
              YesNo[infotbl(cap_ptr+4) and $10=$10]);

            WriteLn('     >4GiB Address Space Supported : ',
              YesNo[infotbl(cap_ptr+4) and $20=$20]);

            WriteLn('     Sideband Addressing Supported : ',
              YesNo[infotbl(cap_ptr+5) and 2=2]);

            Write  ('     Maximum Command Queue Length : ',infotbl(cap_ptr+7)+1,' byte');
            if infotbl(cap_ptr+7)=0 then WriteLn else WriteLn('s');


            { Command register }

            Write  ('     AGP Speed Selected : ');
            if infotbl(cap_ptr+8) and 8=8 then
              begin
                if infotbl(cap_ptr+8) and 1=1 then Write('4x ');
                if infotbl(cap_ptr+8) and 2=2 then Write('8x ');
                if infotbl(cap_ptr+8) and 4=4 then Write('Unknown Speed Reported! ');
                if infotbl(cap_ptr+8) and 2=0 then Write('None Selected');
              end
            else
              begin
                if infotbl(cap_ptr+8) and 1=1 then Write('1x ');
                if infotbl(cap_ptr+8) and 2=2 then Write('2x ');
                if infotbl(cap_ptr+8) and 4=4 then Write('4x ');
                if infotbl(cap_ptr+8) and 7=0 then Write('None Selected');
              end;
            WriteLn;

            WriteLn('     FW Transfers Enabled : ',
              YesNo[infotbl(cap_ptr+8) and $10=$10]);

            WriteLn('     >4GiB Address Space Enabled : ',
               YesNo[infotbl(cap_ptr+8) and $20=$20]);

            Write  ('     AGP Enabled : ');
            if infotbl(cap_ptr+9) and 1=1 then
              WriteLnColor('Yes',LightGreen)
            else
              WriteLnColor('No',LightRed);

            WriteLn('     Sideband Addressing Enabled : ',
              YesNo[infotbl(cap_ptr+9) and 2=2]);

            Write  ('     Current Command Queue Length : ',infotbl(cap_ptr+11)+1,' byte');
            if infotbl(cap_ptr+11)=0 then WriteLn else WriteLn('s');
          end;


        03 :
          begin
            WriteLn('   Vital Product Data Capability');
          end;


        04 :
          begin
            WriteLn('   Slot Identification Capability');

            Write  ('     This is ');
            if infotbl(cap_ptr)+2 and $20=0 then Write('not ');
            WriteLn('a parent bridge');

            Write  ('     Number of slots on secondary side of this bridge : ');
            WriteLn(infotbl(cap_ptr+2) and $1f);

            WriteLn('Chassis Number : ',infotbl(cap_ptr+3));
          end;


        05 :
          begin
            WriteLn('   Message Signalled Interrupt Capability');
            Write  ('     MSI is ');
            if infotbl(cap_ptr+2) and 1=1 then WriteLn('enabled') else WriteLn('disabled');

            Write  ('     MSI function can generate ');
            if infotbl(cap_ptr+2) and 128=128 then WriteLn('64') else Write('32');
            WriteLn('-bit addresses');
          end;


        06 :
          begin
            WriteLn('   CompactPCI Hot-Swap Capability');
          end;


        07 :
          begin
            WriteLn('   PCI-X Capability');

{ type 1 }
            if infotbl($e) and $7f=1 then
              begin
                Write  ('     Secondary AD Bus size is ');
                if infotbl(cap_ptr+2) and 1=1 then Write('64') else Write('32');
                WriteLn('-bit');

                Write  ('     Secondary AD Bus Maximum Speed in PCI-X mode is ');
                if infotbl(cap_ptr+2) and 2=2 then Write('133') else Write('66');
                Writeln('MHz');

                Write  ('     Seconday AD Bus Current Speed : ');
                case Word((infotbl(cap_ptr+3) and 1) shl 8 + (infotbl(cap_ptr+2) and $c0) shr 6) of
                  0 : Write( '33');
                  1 : Write( '66');
                  2 : Write('100');
                  3 : Write('133');
               4..7 : Write('Unknown!');
                end;

              end;

{ type 0 and 1 }
            Write  ('     Primary AD Bus size is ');
            if infotbl(cap_ptr+6) and 1=1 then Write('64') else Write('32');
            Writeln('-bit');

            Write  ('     Primary AD Bus Maximum Speed in PCI-X mode is ');
            if infotbl(cap_ptr+6) and 2=2 then Write('133') else Write('66');
            WriteLn('MHz');
          end;


        09 :
          begin
            WriteLn('   Vendor-Dependant Capability');
          end;

        $0a :
          begin
            WriteLn('   Debug Port Capability');
          end;

        $0b :
          begin
            WriteLn('   CompactPCI Resource Control Capability');
          end;

        $0c :
          begin
            WriteLn('   PCI Hot-Plug Capability');
          end;

      else
            WriteLn('   Unknown Capability (Code ',wrhex(infotbl(cap_ptr)),'h)!!');
      end;

      cap_ptr:=infotbl(cap_ptr+1);

    until cap_ptr=0;
  end;


procedure print_decodestatus(const status:word);
  begin
    Write(wrhexw(status),'h');
    if status<>0 then
      begin
        first:=true;
        Write(' (');
        (* Bit 3?..0 reserved *)
        if Odd(status shr  3) then printstatus('Signalled Interrupt');
        if Odd(status shr  4) then printstatus('Has Capabilities List');
        if Odd(status shr  5) then printstatus('Supports 66MHz');
        if Odd(status shr  6) then printstatus('Has UDF');
        if Odd(status shr  7) then printstatus('Supports Back-To-Back Trans.');

        if Odd(status shr  8) then printstatus('Data parity Error Detected');
        (* timing 2/4*)
        if Odd(status shr 11) then printstatus('Signalled Target Abort');
        if Odd(status shr 12) then printstatus('Received Target Abort');
        if Odd(status shr 13) then printstatus('Received Master Abort');
        if Odd(status shr 14) then printstatus('Signalled System Error'); (* 'Received System Error' *)
        if Odd(status shr 15) then printstatus('Detected Parity Error');

        case (status shr 9) and 3 of
          0 : printstatus('Fast Timing');
          1 : printstatus('Medium Timing');
          2 : printstatus('Slow Timing');
          3 : printstatus('Unknown Timing');
        end;
        Write(')');

      end;
  end;

procedure print_decodebridgecontrol(const control:word;const devtype:byte);
  begin
    Write(wrhexw(control),'h ');
    if control<>0 then
      begin
        first:=true;
        Write('(');
        if Odd(control shr  0) then printstatus('parity detection');
        (* reserved, cardbus *)
        if Odd(control shr  1) then printstatus('CSERR forwarded');
        if Odd(control shr  2) then printstatus('ISA mapping');
        if Odd(control shr  3) then printstatus('VGA mapping');
        (* reserved *)
        if Odd(control shr  5) then printstatus('master abort mode');
        if Odd(control shr  6) then printstatus('bus is in RESET');
        case devtype of
          2:
            begin
              if Odd(control shr  7) then printstatus('ints routed by ExCA')
              else                        printstatus('ints routed by PCI');
            end;
        else
          if Odd(control shr  7) then printstatus('back-to-back transactions');
        end;
        (* cardbus:
        if Odd(control shr  8) then printstatus('memory 0 prefetchable');
        if Odd(control shr  9) then printstatus('memory 1 prefetchable'); *)
        if Odd(control shr 10) then printstatus('write post enable');
        (* 15..11 reserved *)
        Write(')');
      end;
  end;

function has_agp_capability:boolean;
  begin
    has_agp_capability:=false;

    case infotbl($e) and $7f of
      0:cap_ptr:=infotbl($34);
      1:cap_ptr:=infotbl($34);
      2:cap_ptr:=infotbl($14);
    else
        cap_ptr:=0;
    end;

    while cap_ptr<>0 do
      if infotbl(cap_ptr)=02 then (* AGP *)
        begin
          (* any speed supported ? *)
          has_agp_capability:=(infotbl(cap_ptr+4) and 7<>0);
          Exit;
        end
      else
        cap_ptr:=infotbl(cap_ptr+1);

  end;


procedure showallinfo;
  var
    x           : byte;
    sub_vendor,
    sub_device  : word;

  procedure show_address(window_num:word;baseaddr,limit:longint;io_type:boolean;prefetch:boolean);
    const
      prefetch_message:array[boolean] of pChar=(', nonprefetchable',', prfetchable');
    begin
      if baseaddr=0 then exit;

      Write(' Address ',window_num,' is a');

      if io_type then
        Write('n I/O Port : ')
      else
        Write(' Memory Address',prefetch_message[prefetch],' : ');


      WriteLn(wrhexl(baseaddr and $fffffffc),'h, limit : ',wrhexl(limit and $fffffffc),'h');
    end;

  begin

    if businfo then
      begin
        Write(' Bus ');
        TextColor(LightCyan);
        Write(bus);
        TextColor(LightGray);
        Write(' (');

{ we crudely assume bus 1 is AGP, unless we see it's cardbus - this is probably wrong but will do for now!! }
{ It's wrong because a pci-pci bridge on a board without AGP could mean that the far side of the bridge
is reported as AGP, when it's not...}

        case bustype[bus] of
          bustype_AGP:
            Write('AGP');
          bustype_CardBus:
            Write('CardBus');
        else (* bustype_PCI_known,bustype_PCI,bustype_none *)
            Write('PCI');
        end;

        Write('), Device Number ');
        TextColor(LightCyan);
        Write(deviceid);
        TextColor(LightGray);
        Write(', Device Function ');
        TextColor(LightCyan);
        Write(func);
        TextColor(LightGray);
        WriteLn;
      end; (* businfo *)


    if installermode then
      showinstallerinfo
    else
      begin

        Write(' Vendor ',wrhexw(infotbl_W(0)),'h ');
        cmpstr:=wrhexw(infotbl_W(0));
        lookupven(false);
        WriteLn;


        Write(' Device ',wrhexw(infotbl_W(2)),'h ');
        cmpstr:=wrhexw(infotbl_W(2));
        lookupdev(High(String));
        WriteLn;


        if not summary then
          begin
            Write(' Command ',wrhexw(infotbl_W(4)),'h');
            first:=true;
            Write(' (');
            if infotbl(4) and   1=1   then printstatus('I/O Access');
            if infotbl(4) and   2=2   then printstatus('Memory Access');
            if infotbl(4) and   3=0   then printstatus('Bus Access Disabled!!');
            if infotbl(4) and   4=4   then printstatus('BusMaster');
            if infotbl(4) and   8=8   then printstatus('Special Cycles');
            if infotbl(4) and  16=16  then printstatus('MemWrite+Invalidate');
            if infotbl(4) and  32=32  then printstatus('VGA Palette Snoop');
            if infotbl(4) and  64=64  then printstatus('Parity Error Response');
            if infotbl(4) and 128=128 then printstatus('Wait Cycles');

            if infotbl(5) and   1=1   then printstatus('System Errors');
            if infotbl(5) and   2=2   then printstatus('Back-To-Back Transactions');
            if infotbl(5) and   4=4   then printstatus('Interrupt Disable');

            Write(')');
            WriteLn;


            Write(' Status ');
            print_decodestatus(infotbl_W(6));
            WriteLn;

            WriteLn(' Revision ',wrhex(infotbl(8)),'h',
                    ', Header Type ',wrhex(infotbl($e)),'h',
                    ', Bus Latency ',wrhex(infotbl($d)),'h');


            Write(' Self test ',wrhex(infotbl($f)),'h (Self test ');
            if infotbl($f) and $80=0 then Write('not ');
            Write('supported');


            if infotbl($f) and $80=$80 then
              begin
                Write(': Completion code ',wrhexb(infotbl($f) and $f),'h - ');
                if infotbl($f) and $0f=0 then
                  WriteColor('OK',LightGreen)
                else
                  WriteColor('Failed!!',LightRed);
              end;
            WriteLn(')');


            if infotbl($c)<>0 then
              WriteLn(' Cache line size ',infotbl($c)*4,' Bytes (',infotbl($c),' DWords)');


            Write(' PCI Class ');

            i:=infotbl($b);
            if i=$ff then
              begin
                Write('FFh ');
                WriteColor('(does not meet any PCI-SIG defined class)',LightGreen);
              end

            else
            if not (i in [Low(PCI_class_names)..High(PCI_class_names)]) then
              WriteColor(wrhexb(i)+'h Unknown!',Yellow)

            else
              begin

                if i in [Low(PCI_class_names)..High(PCI_class_names)] then
                  WritePCColor(PCI_class_names[i],Yellow);

                Write(', type ');


                found:=false;

                for i:=Low(pci_class_array) to High(pci_class_array) do
                  begin
                    if  (pci_class_array[i].class_  =infotbl($b))
                    and (pci_class_array[i].subclass=infotbl($a))
                    and (pci_class_array[i].progif  =infotbl($9)) then
                      begin
                        found:=true;
                        WritePCColor(PCI_class_array[i].cname,Yellow);
                        Break;
                      end;
                  end;


                if not found then
                  begin
                    for i:=Low(pci_class_array) to High(pci_class_array) do
                      if  (pci_class_array[i].class_  =infotbl($b))
                      and (pci_class_array[i].subclass=infotbl($a))
                      and (pci_class_array[i].progif  =0          ) then
                        begin
                          found:=true;
                          WritePCColor(PCI_class_array[i].cname,Yellow);
                          (* programming interface for storage(1)/ide(1) is:
                              7 bus mastering (read-only)
                              6-4  reserved (read-only)
                              3    secondary IDE mode bit is writable (read-only)
                              2    secondary IDE mode (0 = legacy, 1 = native)
                              1    primary IDE mode bit is writable (read-only)
                              0    primary IDE mode (0 = legacy, 1 = native) *)

                          if (infotbl($b)<>1) or (infotbl($a)<>1) then
                            begin
                              Write(', programming interface ');
                              WriteColor(wrhex(infotbl($9)),Yellow);
                            end;

                          Break;
                        end;
                  end;


                if not found then
                  WriteColor('Unknown!',LightRed);

              end;


            WriteLn;
          end;



        if not summary then
          begin

{ look for generic PCI IDE controller & decode it's info, if present }
            if (infotbl($b)=01) and (infotbl($a)=01) then
              begin
                WriteLn(' PCI EIDE Controller Features :');
                Write  ('   BusMaster EIDE is ');
                if infotbl($9) and $80=0 then
                  WriteColor('NOT ',LightRed);
                WriteLn('supported');

                Write  ('   Primary   Channel is ');
                if infotbl($9) and 1=0 then
                  begin
                    WriteLn('at I/O Port 01F0h and IRQ 14');
                    if infotbl($3c)<>14 then Inc(irqmap[14]);
                  end
                else
                  WriteLn('in native mode at Addresses 0 & 1');

                Write('   Secondary Channel is ');
                if infotbl($9) and 4=0 then
                  begin
                    WriteLn('at I/O Port 0170h and IRQ 15');
                    if infotbl($3c)<>15 then Inc(irqmap[15]);
                  end
                else
                  WriteLn('in native mode at Addresses 2 & 3');
              end;

          end
        else
          begin
{ summary mode: pick up IRQs only }
            if (infotbl($b)=01(*Storage*)) and (infotbl($a)=01(*IDE*)) then
              begin
                if (infotbl($9) and 1=0) and (infotbl($3c)<>14) then Inc(irqmap[14]);
                if (infotbl($9) and 4=0) and (infotbl($3c)<>15) then Inc(irqmap[15]);
              end;
          end;




{ if type 0 table & if Subsystem ID exists, display and scan file for match }
        case infotbl($e) and $7f of
         0:
           begin
             sub_vendor:=infotbl_W($2c);
             sub_device:=infotbl_W($2e);
           end;
         2:
           begin
             sub_vendor:=infotbl_W($40);
             sub_device:=infotbl_W($42);
           end;
        else
             sub_vendor:=0;
             sub_device:=0;
        end;

        if (sub_vendor<>0) or (sub_device<>0) then
          begin

{ subsystem ID }

            Write(' Subsystem ID ',wrhexw(sub_device),
                                   wrhexw(sub_vendor),'h');
            cmpstr:=wrhexw(sub_device)+wrhexw(sub_vendor);


            genssid:=    (sub_vendor=infotbl_W(0))
                     and (sub_device=infotbl_W(2));

            oemidnum:='';
            oemidstr:='';
            bogusid:=false;


            failed:=true;
            userev:=true;
            while (not Eof2(f)) and (failed) do
              begin

{!!}              if userev then vstr:=revchk
                else ReadLn2(f,vstr);
                userev:=false;

                if vstr='' then
                  vstr:=';';

{ OEM Vendor ID }
                case vstr[1] of
                  'O':
                    if StrLComp(@vstr[3],@cmpstr[5],4)=0 then
                      begin
                        oemidstr:=Copy(vstr,8,Length(vstr)); { closest match }
                        oemidnum:=Copy(vstr,3,4); { matching vendor name }
                      end;

                  'S':
                    if StrLComp(@vstr[3],@cmpstr[1],4)=0 then
                      if oemidnum<>'' then
                        begin
                          oemidstr:=Copy(vstr,8,Length(vstr));
                          TextColor(Yellow);
                          Write(' '+oemidstr);
                          if genssid then
                            WriteColor(' (Generic ID)',LightCyan);
                          WriteLn;
                          failed:=false;
                          TextColor(LightGray);
                        end;


{ Oddball 8 digit entry }
                  'X':
                    if StrLComp(@vstr[3],@cmpstr[1],8)=0 then
                      begin
                        oemidnum:=Copy(vstr,7,4); { matching vendor name }
                        bogusid:=true;
                        TextColor(Yellow);
                        Write(' ',Copy(vstr,12,Length(vstr)));
                        if genssid then
                          WriteColor(' (Generic ID)',LightCyan);
                        WriteLn;
                        failed:=false;
                        TextColor(LightGray);
                      end;

                  { remember to ignore comment lines here also!!! }
                  ';':;

                else
                  Break;
                end;

              end; (* while not eof/not failed *)


            if failed then
              begin

                if oemidstr<>'' then
                  begin
                    TextColor(Yellow);
                    Write(' ',oemidstr);
                    TextColor(White);
                    Write(' (Guess Only!)');
                    TextColor(LightGray);
                  end
                else
                  WriteColor(' Unknown',LightRed);

                if genssid then
                  WriteColor(' (Generic ID)',LightCyan);

                WriteLn;
                TextColor(LightGray);

              end;


{ subsystem vendor }
            Write(' Subsystem Vendor ',wrhexw(sub_vendor),'h');

            if bogusid then
              begin
                TextColor(White);
                WriteLn(' Known Bad Subsystem ID - no Vendor ID Available');
                TextColor(LightGray);
              end
            else
              begin
                if oemidnum<>'' then cmpstr:=oemidnum
                else cmpstr:=wrhexw(sub_vendor);

                Close2(f);            { get back to start of file, as the}
                Reset2(f);            { subsys vendor may be higher up...!}

                failed:=true;

                while not Eof2(f) do
                  begin
                    ReadLn2(f,vstr);
                    if (vstr[1]='V') and (StrLComp(@vstr[3],@cmpstr[1],4)=0) then
                      begin
                        TextColor(Yellow);
                        WriteLn(' ',Copy(vstr,8,Length(vstr)));
                        failed:=false;
                        TextColor(LightGray);
                        Break;
                      end;
                  end;
              end;

            if failed then
              begin
                TextColor(LightRed);
                WriteLn(' Unknown');
                TextColor(LightGray);
              end;

          end; (* subsystem id available *)


{ always }
        Close2(f);



        if not summary then
          begin

{ type 0 header = 6 entries, type 1 = 2, type 2 = skip  }
            case infotbl($e) and $7f of
              0:pp:=6;
              1:pp:=1;
            else
                pp:=0;
            end;

            nn:=0;
            while nn<pp do
              begin
                addr:=infotbl_L($10+(nn*4));
                if addr<>0 then
                  begin

                    Write(' Address ',nn,' is a');

                    if addr and 1=1 then
                      begin
                        Write('n I/O Port : ',wrhexl(addr  and $fffffffc),'h');
                      end
                    else
                      begin
                        Write(' Memory Address');
                        case (addr shr 1) and $3 of
                          0:Write(' (anywhere in 0-4GiB');
                          1:Write(' (below 1MiB');
                          2:Write(' (anywhere in 64-bit space');
                          3:Write(' (reserved');
                        end;
                        if addr and 8=8 then Write(', Prefetchable) : ')
                        else Write(') : ');

                        (* 64 bit needs next addr *)
                        if ((addr shr 1) and $3)=2 then
                          begin
                            Write(wrhexl(infotbl_L($10+(nn*4)+4)));
                            Inc(nn);
                          end;

                        Write(wrhexl(addr and $fffffff0)+'h');
                      end;

                    { size the register ?? }
                    WriteLn;
                  end;
                Inc(nn);
              end;

            case (infotbl($e) and $7f) of
              $00:
                begin {other}
                  addr:=infotbl_L($28);
                  if (addr<>0) and (addr<>-1) then
                    begin (* Card Information Structure *)
                      Write(' CIS can read from ');
                      case addr and 7 of
                        0:Write('device specific configuration space');
                        1..6:Write('register ',addr and 7-1,' mapped memory');
                        7:Write('expansion ROM image ',addr shr 28);
                      end;
                      Write(', offset ');
                      addr:=addr and $0ffffff0;
                      if addr<=$ff then
                        Write(wrhex(addr))
                      else
                      if addr<=$ffff then
                        Write(wrhexw(addr))
                      else
                        Write(wrhexl(addr));
                      WriteLn('h');
                    end;
                end;
              $02:
                begin {cardbus bridge}
                  WriteLn(' CardBus Socket/ExCa base memory address is ',wrhexl(infotbl_L($10)),'h');
                  show_address(0,infotbl_L($1c),infotbl_L($20),false,Odd(infotbl_W($3e) shr 8));
                  show_address(1,infotbl_L($24),infotbl_L($28),false,Odd(infotbl_W($3e) shr 9));
                  show_address(2,infotbl_L($2c),infotbl_L($30),true ,false);
                  show_address(3,infotbl_L($34),infotbl_L($38),true ,false);
                end;
            end;

          end;

{ all header types - list IRQ, if present }
        if (infotbl($3c) in [1..254]) or (infotbl($3d)<>0) then
          begin
            Write(' System IRQ ');
            if infotbl($3c) in [0,255] then
              WriteColor('(disabled)',DarkGray)
            else
              Write(infotbl($3c));
            Write(', INT# ');
            if infotbl($3d)=0 then WriteColor('-',DarkGray)
            else Write(chr(infotbl($3d)+Ord('@')));
            WriteLn;
            Inc(irqmap[infotbl($3c)]);
          end;




        if not summary then
          begin


{ type 0,1 header - List ExpROM, if present }
            case infotbl($e) and $7f of
              0:lb:=$30;
              1:lb:=$38;
            else
                lb:=0;

            end;

            if lb<>0 then
              begin
                org_rom_addr:=infotbl_L(lb);
                write_dword(deviceid,func,bus,lb,$fffffffe);
                reset_infotbl_cache;

                (*$IFDEF OLD*)
                if (infotbl(lb+3)<>0) or (infotbl(lb+2)<>0) or (infotbl(lb+1) and $f8<>0) then
                  begin
                    case infotbl(lb+1) of
                      $f8 : romsize:=2;       (* xx xx f8 00 *)
                      $f0 : romsize:=4;
                      $e0 : romsize:=8;
                      $c0 : romsize:=16;
                      $80 : romsize:=32;      (* xx ?? 00 00 *)
                      $00 :
                            case infotbl(lb+2) of
                              $ff : romsize:=64;
                              $fe : romsize:=128;
                              $fc : romsize:=256;
                              $f8 : romsize:=512;
                              $f0 : romsize:=1024;
                              $e0 : romsize:=2048;
                              $c0 : romsize:=4096;
                              $80 : romsize:=8192;
                              $00 : romsize:=16384; { largest possible = 16MiB }
                            end;
                    end;
                (*$ENDIF OLD*)
                (* bit 31..11=start address,10..1=reserved 0=enable *)
                romsize:=((not infotbl_L(lb)+1) shr 11) shl 1;
                write_dword(deviceid,func,bus,lb,org_rom_addr);
                reset_infotbl_cache;
                if romsize<>0 then
                  begin
                    Write(' Expansion ROM of ');
                    if romsize>1000 then
                      Write(romsize/1024:2:0,' MiB')
                    else
                      Write(romsize,' KiB');
                    Write(' decoded by this card, ');
                    if Odd(org_rom_addr) then
                      Write('at address ',wrhexl(org_rom_addr and $fffff800))
                    else
                      WriteColor('currently disabled',DarkGray);
                    WriteLn;
                  end;
              end;


          end; (* summary *)



{ type 1 header only - List bridge info }
        if not summary then
          begin

            if infotbl($e) and $7f=1 then
              begin
                writeln(' PCI Bridge Information:');
                Write  ('   Primary Bus Number ',infotbl($18),', Secondary Bus Number ',infotbl($19));
                if infotbl($19)<=bus then
                  begin
                    TextColor(LightRed);
                    Write(' (invalid)');
                    TextColor(LightGray);
                  end;
                WriteLn(', Subordinate bus number ',infotbl($1a));

{ seconday bus command }
                first:=true;
                Write('   Secondary Bus Command ');
                print_decodebridgecontrol(infotbl_W($3e),1);
                WriteLn;

{ secondary bus status }
                Write('   Secondary Bus Status  ');
                print_decodestatus(infotbl_W($1e));
                WriteLn;

{ latency }
                WriteLn('   Secondary Bus Latency ',wrhex(infotbl($1b)),'h');

{ I/O port range passed by bridge }
                if infotbl_W($1c)<>0 then
                  begin
                    Write('   I/O Port Range Passed to Secondary Bus : ');

{ 16-bit }
                    if infotbl($1c) and $f=0 then
                      begin
                        if infotbl($1d)<infotbl($1c) then
                          WriteLn('None')
                        else
                          WriteLn(wrhexb(infotbl($1c) shr 4),'000h to ',
                                  wrhexb(infotbl($1d) shr 4),'FFFh (16-bit)');
                      end;

{ 32-bit }
                    if infotbl($1c) and $f=1 then
                      begin
                        if infotbl_W($32) < infotbl_W($30) then
                          WriteLn('None')
                        else
                          WriteLn(wrhexw(infotbl_W($30)),wrhexb(infotbl($1c) shr 4),'000h to ',
                                  wrhexw(infotbl_W($32)),wrhexb(infotbl($1d) shr 4),'FFFh (32-bit)');
                      end;
                  end;


{ memory range passed by bridge }
                WriteLn('   Memory   Range Passed to Secondary Bus : ',
                        wrhexw((infotbl_W($20)) and $fff0),'0000h to ',
                        wrhexw((infotbl_W($22)) or  $000f),'FFFFh');



 { optional: prefectchable memory range passed by bridge }
                if (infotbl_W($26)<>0) and (infotbl_W($24)<>0) then
                  WriteLn('   Prefectchable Memory Range Passed to Secondary Bus : ',
                          wrhexw(infotbl_W($24) and $fff0),'0000h to ',
                          wrhexw(infotbl_W($26) or  $000f),'FFFFh');


              end; (* header type 1 *)
          end; (* not summary *)



{ type 2 header only - List bus numbers etc }

        if not summary then
          begin
            if infotbl($e) and $7f=2 then
              begin
                Write(' PCI bus number ',infotbl($18),', CardBus bus number ',infotbl($19));

                if  (infotbl($19)<>0)
                and (bustype[infotbl($19)]<>bustype_CardBus)
                and (bustype[infotbl($19)]<>bustype_PCI_known) then
                  begin
                    TextColor(LightRed);
                    case bustype[infotbl($19)] of
                      bustype_PCI:
                        Write(' (is already PCI bus)');
                      bustype_AGP:
                        Write(' (is already AGP bus)');
                    end;
                    TextColor(LightGray);
                  end;

                WriteLn(', Subordinate bus number ',infotbl($1a));
                WriteLn(' CardBus latency ',wrhex(infotbl($1b)),'h');

                Write('   Secondary Bus Command ');
                print_decodebridgecontrol(infotbl_W($3e),2);
                WriteLn;

                Write('   Secondary Bus Status  ');
                print_decodestatus(infotbl_W($16));
                WriteLn;
              end;
          end;


{ explore the capabilities list, if present }
        if not summary then
          begin
            if (infotbl(6) and $10=$10) then docapdecode;
          end;


{ do a hex-dump, if requested }
        if dumpregs then
          begin
            WriteLn;
            WriteLn(' Hex-Dump of device configuration space follows:');
            for i:=0 to 15 do
              begin
                Write('  ',wrhexw(i*16),'  ');
                for j:=i*16 to i*16+15 do
                  Write(wrhex(infotbl(j)),' ');
                Write('   ');
                for j:=i*16 to i*16+15 do
                  if infotbl(j)<32 then
                    Write('.')
                  else
                    Write(chr(infotbl(j)));
                WriteLn;
              end;
          end;

        WriteLn;
      end; (* not installermode *)

  end; (* showallinfo *)



procedure search_file(var filefound:string;const fname,pathvar:PChar);
  var
    filefound_p:array[0..260] of char;
  begin
    WinDos.FileSearch(filefound_p,fname,WinDos.GetEnvVar(pathvar));
    filefound:=StrPas(filefound_p);
  end;

var
  pins_used     :boolean;
  multifunction :boolean;
  sfunction     :word;
  count         :word;
  h             :byte;
  link_use      :array[0..255] of set of byte;
  irq_usage_documented: array[0..31] of boolean;


begin
  {$IfDef OS2}
  Open_oemhlp;
  os2_read_bios;
  {$EndIf}

  showhelp:=false;
  businfo:=false;
  dorouting:=true;
  dopcirouting:=false;
  dumpregs:=false;
  usebios:=true;
  summary:=false;
  installermode:=false;
  {$IfDef VirtualPascal}
  vga50:=SysCtrlSelfAppType<2;
  {$Else}
  vga50:=true;
  {$EndIf}

{ the following hack permits MS-DOS display output redirection to work }
  if ioredirected then
    begin
      Assign(output,'');
      Rewrite(output);
    end
  else
{ code to do page pausing }
    begin
      ClrScr;
      install_pager;
    end;



  for i:=Low(irqmap) to High(irqmap) do irqmap[i]:=0;
  failed:=true;


  if ParamCount>0 then
    for i:=1 to ParamCount do
      begin
        cmdstr:=ParamStr(i);
        for j:=1 to Length(cmdstr) do
          cmdstr[j]:=UpCase(cmdstr[j]);
        if (Length(cmdstr)=Length('-?')) and (cmdstr[1] in ['+','-','/']) then
          case cmdstr[2] of
            'H':usebios:=false;
            'D':dumpregs:=true;
            'S':summary:=true;
            'T':dorouting:=false;
            'P':dopcirouting:=true;
            'B':businfo:=true;
            'I':installermode:=true;
            '5':vga50:=false;
          else
                showhelp:=true;
          end
        else
          showhelp:=true;

        if showhelp then
          begin
            if vga50 then
              TextMode(Co80);
            WriteLn(' Help for PCI  (Version ',revision,')');
            TextColor(DarkGray);
            WriteLn('');
            TextColor(LightGray);
            WriteLn;
            WriteLn('Usage: PCI [-H] [-D] [-S] [-T] [-B] [-P] [-?]   [] indicates optional parameter');
            WriteLn;
            WriteLn;
            WriteLn('-H : Use direct hardware access (instead of the BIOS) to retrieve PCI Info');
            WriteLn('     May be required for accurate reporting on Intel 430FX chipset+Award BIOS');
            WriteLn('-D : Do a hex-dump of each device''s entire configuration space');
            WriteLn('-S : Create a brief, summary report only; only devices and IRQs listed');
            WriteLn('-T : Disable the test ROM IRQ Routing Table function');
            WriteLn('-B : Enable display of the Bus, Device & Function information');
            WriteLn('-P : Enable display of PCI slot routing data');
            WriteLn('-I : Installer mode: produce raw data dump (for use with auto-setup programs)');
            WriteLn('-? : Displays this help screen!');
            WriteLn;
            WriteLn('PCI Supports generating reports to a file or printer using MS-DOS pipes; i.e.');
            WriteLn;
            WriteLn('  PCI -D > REPORT.TXT  (Save report to file),  PCI > LPT1:  (Print report)');
            WriteLn;
            WriteLn('PCI is written by Craig Hart, and is released as freeware, with no restictions');
            Write('on use or copying. Visit ');
            WriteColor('http://members.hyperlink.net.au/~chart ',LightCyan);
            WriteLn('for updates to');
            WriteLn('the program and the PCI Database file PCIDEVS.TXT');
            Halt(10);
          end;
      end;


{ fix up conflicting commandline switches }

  if installermode then
    begin
      dorouting:=false;
      dopcirouting:=false;
      dumpregs:=false;
      businfo:=false;
      summary:=false;
      vga50:=false;
    end;

  if summary then
    begin
      dumpregs:=false;
      dopcirouting:=false;
      dorouting:=false;
    end;

  if not ioredirected then
    if vga50 then
      TextMode(Co80+Font8x8);


  if not installermode then
    begin

      WriteLn('Craig Hart''s PCI+AGP bus sniffer, version ',revision,', freeware made in 1996-2004.');

      { calculate datafile searchpath: exefile path,... }
      pcidevs_path:=ParamStr(0);
      while (not (pcidevs_path[Length(pcidevs_path)] in ['\','/'])) and (pcidevs_path<>'') do
        Dec(pcidevs_path[0]);

      pcidevs_txt:=FSearch('pcidevs.txt',pcidevs_path);

      if pcidevs_txt='' then
        search_file(pcidevs_txt,'pcidevs.txt','PATH');

      if pcidevs_txt='' then
        search_file(pcidevs_txt,'pcidevs.txt','DPATH');


      if pcidevs_txt<>'' then
        begin
          Assign2(f,pcidevs_txt);
          {$i-}
          Reset2(f);
          {$i+}
        end;

      if (IOResult<>0) or (pcidevs_txt='') then
        begin
          WriteLn('PCI Halted:');
          WriteLn;
          WriteLn('Sorry, I cannot locate my PCIDEVS.TXT datafile!!!');
          Halt(10);
        end;
      Close2(f);
      {$i+}
    end;



  if test8086<2 then
    begin
      WriteLn('PCI Halted:');
      WriteLn;
      WriteLn('The PC Must be at least a 386 to possibly have a PCI or AGP bus!');
      Halt(1);
    end;

{ Look for PCI BIOS }
  pci_present_test;

  if failed then
    begin
      WriteLn('PCI Halted:');
      WriteLn;
      WriteLn('No PCI BIOS was detected! (NB: I don''t work under Windows NT & Windows 2000!)');
      Halt(2);
    end;



{ OK, we have PCI... do our stuff.. }


  if not installermode then
    begin

      (*WriteLn(' Craig Hart''s PCI+AGP bus sniffer, version ',revision,', freeware made in 1996-2004.');*)
      WriteLn;
      Write('PCI BIOS Version ',PCIverhi,'.',wrhex(PCIverlo),' found!');

      if summary then WriteLn('                                  (Summary Report)') else

      WriteLn;
      WriteLn('Number of PCI Busses : ',PCI_hibus+1);
      Write('PCI Characteristics  : ');
      if PCIchar and  1=1  then Write('Config Mechanism 1 ')
      else usebios:=true; { must use BIOS if no cfg mech 1 supported }
      if PCIchar and  2=2  then Write('Config Mechanism 2 ');
      if PCIchar and 16=16 then Write('Special Cycle Mechanism 1 ');
      if PCIchar and 32=32 then Write('Special Cycle Mechanism 2 ');
      WriteLn;
      WriteLn;
      Write('Searching for PCI Devices using ');
      if usebios then
        {$IfDef OS2}
        WriteLn('the OEMHLP$ driver')
        {$Else}
        WriteLn('the System BIOS')
        {$EndIf}
      else
        WriteLn('Configuration Mechanism 1');
      WriteLn;

    end; (* not installermode *)

  for bus:=Low(bustype) to High(bustype) do
    if bus<=pci_hibus then
      bustype[bus]:=bustype_PCI_known   { fix bugs for 440LX chipset, 2 PCI buses, AGP=1 bus! ?}
    else
      bustype[bus]:=bustype_None;


  for bus:=Low(bustype) to High(bustype) do

    if (bustype[bus]<>bustype_None) or (bus<=pci_hibus) then

      for deviceid:=0 to $1f do
        begin

          for func:=0 to 7 do
            begin

              set_lookup_device(deviceid,func,bus);
{don't try to read cfg-space of non-existant devices: hangs some chipsets!}
              if infotbl_W(0)<>$ffff then
                begin

{ remember CardBus stuff for later; skip if far bus=0 (i.e. unconfigured) }
                  if (infotbl($19)>bus) then (* bus<>0 and not already known bus *)
                    case infotbl($e) and $7f of
                      (* 0:other *)
                      1: (* PCI-PCI (+AGP) *)
                        begin
                          (* maybe look for "AGP" in the database entry ? *)
                          if (bus=0) and ((infotbl($3e) and $08)=$08) then
                            bustype[infotbl($19)]:=bustype_AGP
                          else
                            bustype[infotbl($19)]:=bustype_PCI;
                          pci_hibus:=Max(pci_hibus,infotbl($19)); (* secondary bus   *)
                          pci_hibus:=Max(pci_hibus,infotbl($1a)); (* subordinate bus *)
                        end;
                      2: (* CardBus *)
                        begin
                          if bustype[infotbl($19)] in [bustype_None,bustype_PCI_known] then
                            bustype[infotbl($19)]:=bustype_CardBus;
                          pci_hibus:=Max(pci_hibus,infotbl($19)); (* secondary bus   *)
                          pci_hibus:=Max(pci_hibus,infotbl($1a)); (* subordinate bus *)
                        end;
                    end;

                  if terminate_request then Halt(9);
                  showallinfo;

{ If not multi-device device, then don't test for func 1-7 as some cards
Incorrectly answer back on all 8 function numbers!!! S3 trio64, for example - stupid!  }
                  if (func=0) and (infotbl($e) and $80=0) then
                    Break; (* func:=7; *)
                end; (* exist *)

            end; (* Function *)

        end; (* DeviceID *)

     (* Bus *)



{
  The following is an experiment with "Get IRQ Routing Info" BIOS function:
  the avid coder is free to un-comment the code and try it out: I couldn't
  make much sense out of the information returned myself!
}



  if dopcirouting then
    begin

      WriteLn;
      WriteLn('PCI slot IRQ mapping information');

      failed:=true;
      FillChar(irqbuff,SizeOf(irqbuff),$00);
      load_irqbuff;

      if not failed then
        begin

          TextColor(LightGreen);
          WriteLn(' PCI slot mapping information read successfully');
          TextColor(LightGray);
          {$IfDef OS2}
          if map_of_IRQ_channels_permanently_dedicated_to_PCI_unknown then
            begin
              TextColor(LightRed);
              WriteLn(' But IRQ permanently dedicated to PCI is unknown, assume none.');
              TextColor(LightGray);
            end;
          {$EndIf OS2}
          WriteLn;


{ hex-dump table }
          if dumpregs then dohexdump;

{}
          WriteLn(' PCI slot IRQ availability listing');
          WriteLn;
          for i:=0 to (len shr 4)-1 do
            with irqbuffR.IRQ_routing_table_entry_Array[i] do
              begin
                WriteLn('  PCI Bus ',PCI_bus_number,', Device ',PCI_device_number shr 3,', Slot ',wrhex(device_slot_number));
                for j:=Low(INTABCD) to High(INTABCD) do
                  listmap(INTABCD[j].IRQ_connectivity_bit_map,'   INT'+Chr(Ord('A')+j)+'# can be connected to IRQs ');
                WriteLn;
              end;
          verify_routing_table(irqbuffR.IRQ_routing_table_entry_Array,len shr 4);
          WriteLn;


{}
          for i:=Low(link_use) to High(link_use) do
            link_use[i]:=[];

          FillChar(irq_usage_documented,SizeOf(irq_usage_documented),false);

          WriteLn(' PCI slot INTx to IRQ-router mappings');
          WriteLn;
          WriteLn('  SLOT BUS DEV  INTA INTB INTC INTD');
          for i:=0 to (len shr 4)-1 do
            with irqbuffR.IRQ_routing_table_entry_Array[i] do
              begin
                if PCI_bus_number=0 then
                  irq_usage_documented[PCI_device_number shr 3]:=true;
                Write('   ',wrhex(device_slot_number),'  ',PCI_bus_number:2,'  ',PCI_device_number shr 3:2);
    {           Write('   ',PCI_device_number and 3);}
                Write('   ');
                (*              1   2   3   4
                  PIIX          $60 $61 $62 $63
                  VT596         1   2   3   5   ??
                  M1543         $48 $c8 $49 $c9
                  SIS5513       $41 $42 $43 $44         *)

                for j:=Low(INTABCD) to High(INTABCD) do
                  Write(' ',wrhex(INTABCD[j].link_value),'  ');

                pins_used:=false;
                sfunction:=0;
                repeat
                  set_lookup_device(PCI_device_number shr 3,
                                    sfunction,
                                    PCI_bus_number);

                  if infotbl_W(0)=$ffff then
                    begin
                      if sfunction=0 then Break;
                    end
                  else
                    begin
                      (* PCI IDE IRQ 15/14 *)
                      if (infotbl($b)=01) and (infotbl($a)=01) and ((infotbl($9) and $5)=0) then
                        (* no pin? *)
                      else
                        if infotbl($3d) in [1..4] then
                          begin
                            if not pins_used then
                              begin
                                WriteLn;
                                pins_used:=true;
                              end;
                            Write('          IRQ:','':2+(infotbl($3d)-1)*5);
                            Write('',infotbl($3c):3,'');
                            Write('':(4-infotbl($3d))*5+2);
                            cmpstr:=wrhexw(infotbl_W(0));
                            lookupven(true);
                            cmpstr:=wrhexw(infotbl_W(2));
                            lookupdev(42);
                            WriteLn;
                            link_use[INTABCD[infotbl($3d)-1].link_value]:=link_use[INTABCD[infotbl($3d)-1].link_value]
                                                                         +[infotbl($3c)];
                          end;
                    end;

                  if sfunction=0 then
                    multifunction:=Odd(infotbl($e) shr 7);

                  Inc(sfunction);
                until (sfunction>7) or (not multifunction);

                if not pins_used then
                  WriteLn('No Device Detected');
              end;

          WriteLn;

          (* check for bad tables: one link value can cause only one IRQ line *)
          for i:=Low(link_use) to High(link_use) do
            begin
              count:=0;
              for h:=1 to 254 do
                if h in link_use[i] then
                  Inc(count);
              if count>1 then
                begin
                  TextColor(LightRed);
                  WriteLn('Error for link value ',wrhex(i),':');
                  Write  ('can not connect to more than one of IRQ ');
                  count:=0;
                  for h:=1 to 254 do
                    if h in link_use[i] then
                      begin
                        if count>0 then Write(',');
                        Write(h);
                        Inc(count);
                      end;
                  WriteLn;
                  TextColor(LightGray);
                end;
            end;


          (* check for bad tables: not listed devices for bus 0*)
          for i:=Low(irq_usage_documented) to High(irq_usage_documented) do
            if not irq_usage_documented[i] then
              begin
                (* only look on first function *)
                set_lookup_device(i,0,0);
                if infotbl_W(0)<>$ffff then
                  if infotbl($3c) in [1..254] then
                    if infotbl($3d) in [1..4] then
                      begin
                        TextColor(LightRed);
                        Write('Table error: Bus=0, Device=',i,' not listed! ');
                        cmpstr:=wrhexw(infotbl_W(0));
                        lookupven(true);
                        cmpstr:=wrhexw(infotbl_W(2));
                        lookupdev(42);
                        TextColor(LightGray);
                        WriteLn;
                      end;
              end;

{}
          listmap(conmap,' IRQ''s dedicated to PCI : ');

        end
      else
        begin
          TextColor(LightRed);
          WriteLn(' Unable to read slot mapping information from PCI BIOS!');
          TextColor(LightGray);
        end;
      WriteLn;

    end; (* dopcirouting *)



{ BIOS IRQ Routing table tests }

    if dorouting then
      showroutinginfo;


{ final summarial IRQ info }

    if not installermode then
      begin
        WriteLn;
        Write('IRQ Summary: ');
        failed:=true;
        disp:=0;
        irqmap[0]:=0;
        irqmap[255]:=0;
        for i:=Low(irqmap) to High(irqmap) do if irqmap[i]>0 then Inc(disp); { count IRQs}
        for i:=Low(irqmap) to High(irqmap) do if irqmap[i]>0 then
          begin
            if failed then
              begin
                if disp=1 then Write('IRQ ')
                else           Write('IRQs ');
              end
            else
              Write(',');
            Write(i);
            failed:=false;
          end;
        if failed then WriteLn('No IRQ''s are used by PCI Devices!')
        else
          begin
            if disp=1 then Write(' is') else Write(' are');
            WriteLn(' used by PCI devices');
          end;

        Write('Shared IRQs: ');
        failed:=true;
        for i:=Low(irqmap) to High(irqmap) do if irqmap[i]>1 then
          begin
            if not failed then Write('             ');
            WriteLn('IRQ ',i,' is shared by ',irqmap[i],' PCI Devices');
            failed:=false;
          end;
        if failed then WriteLn('There are no shared PCI IRQs');

      end; (* not installermode *)

  {$IfDef OS2}
  Close_oemhlp;
  {$EndIf}
end.
